''' Pythonic representation of the Constant GAM '''
from martepy.marte2.gam import MARTe2GAM
from martepy.marte2.qt_functions import addOutputSignalsSection

class ConstantGAM(MARTe2GAM):
    """This block allows you to define multiple constant signals
    with multiple dimensions and elements. You can also adjust the
    constant values during runtime via messages
    """
    def __init__(self,
                    configuration_name: str = 'Constants',
                    output_signals: list = [],
                    input_signals = []
                ):
        #assert all(('Default' in d['MARTeConfig'] for n, d in output_signals))
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'ConstantGAM',
                output_signals = output_signals,
                input_signals = []
            )

    def writeGamConfig(self, _):
        ''' Write an empty configuration '''

    def serialize(self):
        ''' Serialize nothing but the base class setup '''
        res = super().serialize()
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the object '''
        super().deserialize(data, hashmap, restore_id)
        return self

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can call
        the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        app_def = mainpanel_instance.parent.API.getServiceByName('ApplicationDefinition')
        datasource = app_def.configuration['misc']['gamsources'][0]
        addOutputSignalsSection(mainpanel_instance, node, default=True, datasource=datasource)


def initialize(factory, plugin_datastore) -> None:
    ''' Initialize the object with the factory '''
    factory.registerBlock("ConstantGAM", ConstantGAM, plugin_datastore)



''' Pythonic representative class of the Logger Datasource '''

from martepy.marte2.datasource import MARTe2DataSource

from martepy.marte2.qt_functions import addSignalsSection

class LoggerDataSource(MARTe2DataSource):
    ''' Pythonic representation of the Logger DataSource '''
    def __init__(self,
                    configuration_name: str = 'LoggerDataSource',
                    input_signals = [],
                    output_signals = []
                ):
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'LoggerDataSource'
            )

    def writeDatasourceConfig(self, config_writer):
        pass

    def serialize(self):
        res = super().serialize()
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        super().deserialize(data, hashmap, restore_id)
        # Now we build up
        return self

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can
        call the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        addSignalsSection(mainpanel_instance, node, type_input=True)



def initialize(factory, plugin_datastore) -> None:
    ''' Initialize our object with the factory '''
    factory.registerBlock("LoggerDataSource", LoggerDataSource, plugin_datastore)



''' The window in the XMARTe2 GUI that loads to configure signals, exists here as is 
launched from the GAM itself. '''

from functools import partial
import os
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
    QDesktopWidget,
    QWidget,
    QSizePolicy,
    QTableWidget,
    QTableWidgetItem,
    QVBoxLayout,
    QHBoxLayout,
    QPushButton,
    QMainWindow
)
from qtpy.QtGui import QIcon

from martepy.marte2.datasource import MARTe2DataSource
from martepy.functions.extra_functions import getname


class SignalWdw(QMainWindow):
    ''' The configurable signal window where the user can configure signals in XMARTe2. '''
    def __init__(self, parent=None, node=None, hasdefault=False, epics=False,
                 samples=False, io='output'):
        if node is None:
            raise ValueError("Node cannot be null")
        super().__init__(parent)
        self.node = node
        self.to_delete = []
        self.epics = epics
        self.samples = samples
        self.io = io
        self.app = node.application.app
        self.setWindowTitle("Signal Configuration")
        # Setup our geometry
        self.setSize()
        self.setCentralWidget(QWidget(self))
        self.vlayout = QVBoxLayout()
        self.default = hasdefault
        self.centralWidget().setLayout(self.vlayout)
        # Section which defines input signals and their options selected
        self.signal_tbl = QTableWidget()
        self.vlayout.addWidget(self.signal_tbl)
        signals = node.outputsb if self.io == 'output' else node.inputsb
        self.signal_tbl.setRowCount(len(signals))
        headers = ["Signal Name", "Datasource", "Type",
                   "NumberOfDimensions", "NumberOfElements", 'Alias']


        self.is_datasource = isinstance(self.node.application.API.toGAM(self.node),
                                                       MARTe2DataSource)
        self.signal_tbl.setColumnCount(7 + self.handleAdditionalHeaders(headers))
        headers += ['Delete']
        self.signal_tbl.setHorizontalHeaderLabels(headers)

        # Connect the itemChanged signal to the handleItemChanged function
        self.signal_tbl.itemChanged.connect(self.handleItemChanged)
        for row_num, signal in enumerate(signals):
            self.createSignalRow(signal, self.is_datasource, row_num)

        self.handleSamples()

        self.handleIo()

        self.defineSaveCancelButtons(self.vlayout)
        self.resize(self.signal_tbl.horizontalHeader().length() + 80, self.height())
        self.show()
        # Center the window on the screen
        self.center()

    def handleSamples(self):
        ''' Handles whether our signal supports the samples attribute '''
        colcount = self.signal_tbl.columnCount()
        if self.samples and self.io == 'input':
            # For a MuxGAM we only want to expose the ability to set the number of samples
            for row in range(self.signal_tbl.rowCount()):
                for i in range(0,colcount):
                    item = self.signal_tbl.item(row, i)
                    item.setFlags(item.flags() & ~Qt.ItemIsEditable)
                item = self.signal_tbl.item(row, colcount-1)
                item.setFlags(item.flags() & ~Qt.ItemIsEditable)

    def handleIo(self):
        ''' Handle whether our signal is an input and therefore the user cannot 
        change the alias '''
        if self.io == 'input':
            # Make column 5 read-only
            for row in range(self.signal_tbl.rowCount()):
                item = self.signal_tbl.item(row, 5)
                if item:
                    item.setFlags(item.flags() & ~Qt.ItemIsEditable)
                if self.is_datasource:
                    item = self.signal_tbl.item(row, 1)
                    if item:
                        item.setFlags(item.flags() & ~Qt.ItemIsEditable)


    def handleAdditionalHeaders(self, headers):
        ''' Handle whether we need additional headers '''
        header_cnt = 0
        if self.default:
            headers += ['Default']
            header_cnt += 1
        if self.epics:
            headers += ["PVName"]
            header_cnt += 1

        if self.samples:
            headers += ['Samples']
            header_cnt += 1
        return header_cnt

    def createSignalRow(self, signal, is_datasource, row_num):
        ''' Given a signal, add it to our table '''
        def getSetKey(key, default_val):
            ''' Check that a key exists and if not, create it '''
            if key not in list(signal[1]['MARTeConfig'].keys()):
                signal[1]['MARTeConfig'][key] = default_val
            return signal[1]['MARTeConfig'][key]

        self.signal_tbl.setItem(row_num, 0, QTableWidgetItem(signal[0]))

        if is_datasource:
            signal[1]['MARTeConfig']['DataSource'] = getname(self.node)
        datasource = getSetKey('DataSource', None)
        sgl_type = getSetKey('Type', None)
        self.signal_tbl.setItem(row_num, 1, QTableWidgetItem(datasource))
        self.signal_tbl.setItem(row_num, 2, QTableWidgetItem(sgl_type))
        # - note in constantGAM's this could produce an issue so be wary
        # about changing the node itself unless this is itself changed!
        dimensions = getSetKey('NumberOfDimensions', '1')
        elements = getSetKey('NumberOfElements', '1')
        self.signal_tbl.setItem(row_num, 3, QTableWidgetItem(dimensions))
        self.signal_tbl.setItem(row_num, 4, QTableWidgetItem(elements))
        # Set the alias
        alias = getSetKey('Alias', signal[0])
        self.signal_tbl.setItem(row_num, 5, QTableWidgetItem(alias))
        colcount = 6
        if self.default:
            default = getSetKey('Default', '{1}')
            self.signal_tbl.setItem(row_num, colcount, QTableWidgetItem(default))
            colcount += 1
        if self.epics:
            pv_name = getSetKey('PVName', signal[0])
            self.signal_tbl.setItem(row_num, colcount, QTableWidgetItem(pv_name))
            colcount += 1
        if self.samples:
            samples = getSetKey('Samples', signal[0])
            self.signal_tbl.setItem(row_num, colcount, QTableWidgetItem(samples))
            colcount += 1

        del_button = QPushButton()
        del_button.setIcon(QIcon(os.path.join('xmarte','qt5','icons','delete.png')))
        del_button.clicked.connect(partial(self.deleteRow, row_num))
        self.signal_tbl.setCellWidget(row_num, colcount, del_button)

    def center(self):
        ''' Centres our window in the screen '''
        # Get the geometry of the main window
        qr = self.frameGeometry()
        # Get the center point of the screen
        cp = QDesktopWidget().availableGeometry().center()
        # Move the center point of the main window to the center point of the screen
        qr.moveCenter(cp)
        # Set the top-left corner of the main window to the top-left corner of the qr geometry
        self.move(qr.topLeft())

    def handleItemChanged(self, item):
        ''' If an item has changed and it is the alias, and an output,
        then change the name as well to match '''
        if item.column() == 0 and self.io == 'output':  # Check if the modified item is in column 1
            row_number = item.row()  # Get the row number of the modified item
            column_1_text = item.text()  # Get the text of the modified item in column 1

            # Modify the item in column 7 of the same row to match the text contents of column 1
            self.signal_tbl.setItem(row_number, 5, QTableWidgetItem(column_1_text))

    def deleteRow(self, row):
        ''' Set the current signal to be deleted unless the window is not saved when closed '''
        self.signal_tbl.removeRow(row)
        socket_list = self.node.outputs if self.io == 'output' else self.node.inputs
        socket_to_delete = socket_list[row]
        self.to_delete += [(row,socket_to_delete)]


    def defineSaveCancelButtons(self, layout):
        """Add the generic save and cancel buttons to a given window layout and use self
        for referencing the given expected function names to connect to the button signals.
        """
        spacer = QWidget(self)
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        layout.addWidget(spacer)

        # Now save and cancel buttons
        buttons = QWidget(self)
        hbox = QHBoxLayout()
        button_spacer = QWidget(self)
        button_spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        hbox.addWidget(button_spacer)

        self.save_button = QPushButton("Save")
        self.save_button.clicked.connect(self.save)
        hbox.addWidget(self.save_button)

        self.cancel_button = QPushButton("Cancel")
        self.cancel_button.clicked.connect(self.cancel)
        hbox.addWidget(self.cancel_button)

        buttons.setLayout(hbox)
        layout.addWidget(buttons)

    def setSize(self,x_pos=0.4,y_pos=0.3,width=0.3,height=0.2):
        """Set the Window Size
        """
        # Set Window Size
        self.screen = self.app.primaryScreen()
        self.size = self.screen.size()
        self.setGeometry(
            int(self.size.width() * x_pos),
            int(self.size.height() * y_pos),
            int(self.size.width() * width),
            int(self.size.height() * height),
        )
        self.setCentralWidget(QWidget(self))

    def save(self):
        ''' Save our changes - quite complex as we handle our buffered deletions and delete
        these from the node '''
        if self.to_delete:
            for socket_to_delete in self.to_delete:
                if self.io == 'output':
                    self.node.outputs = [a for a in self.node.outputs if
                                        not a == socket_to_delete[1]]
                else:
                    self.node.inputs = [a for a in self.node.inputs if
                                        not a == socket_to_delete[1]]
                socket_to_delete[1].delete()
                signal_list = self.node.outputsb if self.io == 'output' else self.node.inputsb
                signal_list.pop(socket_to_delete[0])
        signals = self.node.outputsb if self.io == 'output' else self.node.inputsb
        for row in range(len(signals)): # pylint:disable=C0200
            new_tuple = None
            name = self.signal_tbl.item(row, 0).text()
            config = {'MARTeConfig':{'DataSource': self.signal_tbl.item(row, 1).text()}}
            config['MARTeConfig']['Type'] = self.signal_tbl.item(row, 2).text()
            config['MARTeConfig']['NumberOfDimensions'] = self.signal_tbl.item(row, 3).text()
            config['MARTeConfig']['NumberOfElements'] = self.signal_tbl.item(row, 4).text()
            config['MARTeConfig']['Alias'] = self.signal_tbl.item(row, 5).text()
            colcount = 6
            if self.default:
                config['MARTeConfig']['Default'] = self.signal_tbl.item(row, colcount).text()
                colcount += 1
            if self.epics:
                config['MARTeConfig']['PVName'] = self.signal_tbl.item(row, colcount).text()
                colcount += 1
            if self.samples:
                config['MARTeConfig']['Samples'] = self.signal_tbl.item(row, colcount).text()
            sockets = self.node.outputs if self.io == 'output' else self.node.inputs
            for edge in sockets[row].edges:
                try:
                    socket_idx = edge.end_socket.node.inputs.index(edge.end_socket)
                    inputsb = edge.end_socket.node.inputsb[socket_idx]
                    inputsb[1]['MARTeConfig']['Alias'] = self.signal_tbl.item(row, 5).text()
                    inputsb[1]['MARTeConfig']['DataSource'] = self.signal_tbl.item(row, 1).text()
                    edge.end_socket.node.updateSocketPositions()
                except AttributeError:
                    pass
            new_tuple = (name,config)
            signals[row] = new_tuple
            sockets[row].label = name
        self.node.updateSocketPositions()
        application = self.node.application
        self.node.resetParameterbar()
        self.node.onDoubleClicked(None)
        app = application.API.buildApplication()
        application.API.errorCheck(app)
        self.close()
        application.newwindow = None

    def cancel(self):
        ''' Close the window without doing anything '''
        application = self.node.application
        self.close()
        application.newwindow = None



'''
This module is a collection of useful functions throughout the GUI aspects
of the MARTe2 Python repository.
'''
import copy
from functools import partial

from PyQt5.QtWidgets import (QComboBox,
                             QHBoxLayout,
                             QLabel,
                             QLayout,
                             QLineEdit,
                             QMessageBox,
                             QPushButton,
                             QSizePolicy,
                             QWidget,
                             QTextEdit)
from martepy.marte2.objects.configuration_database import MARTe2ConfigurationDatabase
from martepy.marte2.objects.message import MARTe2Message
from martepy.marte2.signal_names_wdw import SignalWdw


def textChangeOut(wdgt, node, default, epics, type_input, datasource=None, # pylint: disable=R0914
        samples=None):
    ''' This function is used to handle when a item has changed for a signal definition
    in a table item. '''
    # Get current number of outputs and add the missing ones or remove.
    if wdgt.text().isdigit():
        if not datasource:  # determine if the node is a datasource itself
            datasource = '' if type_input else node.configuration_name
        orig_sig = node.inputsb if type_input else node.outputsb
        existing_node_names = [a[0] for a in orig_sig]
        existing_app_names = node.application.API.getAliases(node.application)
        name = 'newsignal'
        default_new_sgl = (
            name, {'MARTeConfig': {
                'DataSource': datasource,
                'Type': 'uint32',
                'NumberOfDimensions': '1',
                'NumberOfElements': '1',
                'Alias': name
            }}
        )
        if default:
            default_new_sgl[1]['MARTeConfig']['Default'] = '{0}'
        if epics:
            default_new_sgl[1]['MARTeConfig']['PVName'] = ''
        if samples:
            default_new_sgl[1]['MARTeConfig']['Samples'] = '1'
        try:
            orig = len(node.inputs) if type_input else len(node.outputs)
            change: int = int(wdgt.text()) - orig
        except ValueError:
            change = 0

        shadow = node.inputs if type_input else node.outputs
        endpoint = node.inputsb if type_input else node.outputsb
        if change > 0:

            position = node.input_socket_position if type_input else node.output_socket_position
            multi_edges = node.input_multi_edged if type_input else node.output_multi_edged
            count = len(shadow) - 1
            for i in range(change):
                unique_name = generateUniqueName(existing_app_names + existing_node_names,
                                                   'newsignal')
                existing_node_names.append(unique_name)
                socket = node.__class__.Socket_class(
                    node=node, index=count + i, position=position,
                    socket_type=0, multi_edges=multi_edges,
                    count_on_this_node_side=len(shadow), is_input=type_input, label=unique_name
                )
                shadow += [socket]
                default_new_sgl = (unique_name, default_new_sgl[1])
                default_new_sgl[1]['MARTeConfig']['Alias'] = unique_name
                endpoint += [copy.deepcopy(default_new_sgl)]

        elif change < 0:
            for _ in range(-change):
                del endpoint[-1]
                node.scene.grScene.removeItem(shadow[-1].grSocket)
                for edge in shadow[-1].edges:
                    edge.remove()
                shadow[-1].delete()
                del shadow[-1]

        node.content.updateDim()

def paraChange(value, node, parameter) -> None:
    ''' Update our parameter when there has been a change, this should be connected
    to line edit/comboedit items that are used in the GUI representation of a node. '''
    def updateParameter(text, node, parameter) -> None:
        if isinstance(node.parameters[parameter], int):
            try:
                node.parameters[parameter] = int(text)
            except ValueError:
                pass
        else:
            node.parameters[parameter] = text
    if isinstance(value, QLineEdit):
        updateParameter(value.text(), node, parameter)
    elif isinstance(value, QComboBox):
        updateParameter(value.currentText(), node, parameter)
    elif isinstance(value, QTextEdit):
        updateParameter(value.toPlainText(), node, parameter)

def addLineEdit(mainpanel_instance, node, lbl_name, para_name, row, col_start):
    ''' Adds a line edit item to a node panel '''
    wgt_label = QLabel(lbl_name)
    wgt_field = QLineEdit(mainpanel_instance)
    wgt_field.setText(str(node.parameters[para_name]))
    wgt_field.textChanged.connect(partial(paraChange, wgt_field, node, para_name))
    mainpanel_instance.configbarBox.addWidget(wgt_label, row, col_start)
    mainpanel_instance.configbarBox.addWidget(wgt_field, row, col_start+1)

def addComboEdit(mainpanel_instance, node, lbl_name, para_name, row, col_start, items):
    ''' Add a combo edit item to a node panel '''
    wgt_label = QLabel(lbl_name)
    wgt_field = QComboBox(mainpanel_instance)
    wgt_field.addItems(items)
    wgt_field.setCurrentText(str(node.parameters[para_name]))
    wgt_field.currentIndexChanged.connect(partial(paraChange, wgt_field, node, para_name))
    mainpanel_instance.configbarBox.addWidget(wgt_label, row, col_start)
    mainpanel_instance.configbarBox.addWidget(wgt_field, row, col_start+1)

def addSignalsSection(mainpanel_instance, node, default=False, epics=False, type_input=False):
    ''' Adds a generic signal selection section to a node panel '''
    lbl_wgt = QLabel("Number of signals: ")
    no_signals_wgt = QLineEdit(mainpanel_instance)
    length = len(node.inputsb) if type_input else len(node.outputsb)
    no_signals_wgt.setText(str(length))
    no_signals_wgt.resize(230, 80)
    no_signals_wgt.textChanged.connect(partial(textChangeOut,
                                               no_signals_wgt,
                                               node,
                                               default,
                                               epics,
                                               type_input))
    mainpanel_instance.configbarBox.addWidget(lbl_wgt, 1, 0)
    mainpanel_instance.configbarBox.addWidget(no_signals_wgt, 1, 1)
    colcount = 2
    if not type_input:
        def runWindow():
            config_signals = SignalWdw(mainpanel_instance, node, default, epics)
            node.application.newwindow = config_signals
            config_signals.show()
        config_btn = QPushButton("Configure Signals")
        config_btn.clicked.connect(runWindow)
        mainpanel_instance.configbarBox.addWidget(config_btn, 1, colcount)
        colcount += 1
    spacerwgt = QWidget(mainpanel_instance)
    spacerwgt.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
    mainpanel_instance.configbarBox.addWidget(spacerwgt, 1, colcount)

def addInputSignalsSection(mainpanel_instance, node, pack=True, samples=False,
                           datasource=None, epics=False):
    ''' Add an input signal selection item to a node panel '''
    lbl_wgt = QLabel("Number of input signals: ")
    no_signals_wgt = QLineEdit(mainpanel_instance)
    length = len(node.inputsb)
    no_signals_wgt.setText(str(length))
    no_signals_wgt.resize(230, 80)
    no_signals_wgt.textChanged.connect(partial(textChangeOut, no_signals_wgt,
                                               node, False, epics, True, datasource, samples))
    mainpanel_instance.configbarBox.addWidget(lbl_wgt, 1, 0)
    mainpanel_instance.configbarBox.addWidget(no_signals_wgt, 1, 1)
    # Default means we're probably a constant GAM
    def runWindow():
        config_signals = SignalWdw(mainpanel_instance, node, False, epics, samples,'input')
        node.application.newwindow = config_signals
        config_signals.show()

    config_btn = QPushButton("Configure Signals")
    config_btn.clicked.connect(runWindow)
    mainpanel_instance.configbarBox.addWidget(config_btn, 1, 2)
    if pack:
        spacerwgt = QWidget(mainpanel_instance)
        spacerwgt.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        mainpanel_instance.configbarBox.addWidget(spacerwgt, 1, 2)

def addOutputSignalsSection(mainpanel_instance, node, start = 0, pack=True, datasource=None,
                            samples=False, default=False, epics=False):
    ''' Add an output signal selection item to a node panel. '''
    # Note if samples is set, we'll start from row 2 for cleanliness
    lbl_wgt = QLabel("Number of output signals: ")
    no_signals_wgt = QLineEdit(mainpanel_instance)
    length = len(node.outputsb)
    no_signals_wgt.setText(str(length))
    no_signals_wgt.resize(230, 80)
    no_signals_wgt.textChanged.connect(partial(textChangeOut, no_signals_wgt,
                                               node, default, epics, False, datasource, samples))
    rowstart = 1
    mainpanel_instance.configbarBox.addWidget(lbl_wgt, rowstart, start)
    mainpanel_instance.configbarBox.addWidget(no_signals_wgt, rowstart, start+1)
    def runWindow():
        config_signals = SignalWdw(mainpanel_instance, node, default, epics, samples)
        node.application.newwindow = config_signals
        config_signals.show()
    config_btn = QPushButton("Configure Signals")
    config_btn.clicked.connect(runWindow)
    mainpanel_instance.configbarBox.addWidget(config_btn, rowstart, start+2)
    if pack:
        spacerwgt = QWidget(mainpanel_instance)
        spacerwgt.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
        mainpanel_instance.configbarBox.addWidget(spacerwgt, rowstart, start+3)

def spacer():
    ''' Creates a generic spacer widget '''
    spacerWgt = QWidget()
    spacerWgt.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
    return spacerWgt

def createLineEdit(layout, label, func, finished=False, space=False):
    ''' Creates the actual line edit item used in node panels. '''
    wgt = QWidget()
    hlayout = QHBoxLayout()
    wgt.setLayout(hlayout)
    lbl = QLabel(label)
    edit = QLineEdit()
    if finished:
        edit.editingFinished.connect(func)
    else:
        edit.textChanged.connect(func)
    hlayout.addWidget(lbl)
    hlayout.addWidget(edit)
    if space:
        hlayout.addWidget(spacer())
    layout.addWidget(wgt)
    return edit

def createComboEdit(layout, label, func, items, space=False):
    ''' Creates the actual combobox edit item used in node panels. '''
    wgt = QWidget()
    hlayout = QHBoxLayout()
    wgt.setLayout(hlayout)
    lbl = QLabel(label)
    edit = QComboBox()
    edit.addItems(items)
    edit.currentIndexChanged.connect(func)
    hlayout.addWidget(lbl)
    hlayout.addWidget(edit)
    if space:
        hlayout.addWidget(spacer())
    layout.addWidget(wgt)
    return edit

def createGridLineEdit(layout, row, label, func, finished=False, space=False):
    ''' Creates a line edit item to fit into a grid layout '''
    lbl = QLabel(label)
    edit = QLineEdit()
    if finished:
        edit.editingFinished.connect(func)
    else:
        edit.textChanged.connect(func)
    layout.addWidget(lbl, row, 0)
    layout.addWidget(edit, row, 1)
    if space:
        layout.addWidget(spacer(), row, 2)
    return edit

def createGridComboEdit(layout, row, label, func, items, space=False):
    ''' Creates a combobox item to fit into a grid layout. '''
    lbl = QLabel(label)
    edit = QComboBox()
    edit.addItems(items)
    edit.currentIndexChanged.connect(func)
    layout.addWidget(lbl, row, 0)
    layout.addWidget(edit, row, 1)
    if space:
        layout.addWidget(spacer(), row, 2)
    return edit

def defineSaveCancelButtons(layout, save_func, cancel_func):
    """Add the generic save and cancel buttons to a given window layout and use self
    for referencing the given expected function names to connect to the button signals.
    """
    # Now save and cancel buttons
    buttons = QWidget()
    hbox = QHBoxLayout()
    button_spacer = QWidget()
    button_spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Minimum)
    hbox.addWidget(button_spacer)

    save_button = QPushButton("Save")
    save_button.clicked.connect(save_func)
    hbox.addWidget(save_button)

    cancel_button = QPushButton("Cancel")
    cancel_button.clicked.connect(cancel_func)
    hbox.addWidget(cancel_button)

    buttons.setLayout(hbox)
    layout.addWidget(buttons)

def setSize(window,app,x_pos=0.4,y_pos=0.3,width=0.3,height=0.2):
    """Set the Window Size
    """
    # Set Window Size
    screen = app.primaryScreen()
    size = screen.size()
    window.setGeometry(
        int(size.width() * x_pos),
        int(size.height() * y_pos),
        int(size.width() * width),
        int(size.height() * height),
    )

def recursivelySetEnabled(layout, enable):
    ''' Recursively enable widgets or disable them '''
    for i in range(layout.count()):
        item = layout.itemAt(i)
        if isinstance(item, QLayout):
            recursivelySetEnabled(item, enable)
        elif isinstance(item.widget(), QWidget):
            item.widget().setEnabled(enable)

def generateUniqueName(existing_names, base_name):
    """Generate a unique name based on the existing names in the list.
    """
    if base_name not in existing_names:
        return base_name

    i = 1
    while True:
        new_name = f"{base_name}{i}"
        if new_name not in existing_names:
            return new_name
        i += 1

def generateUniqueGamName(existing_names):
    """Generate a unique name based on the existing names in the list.
    """
    i = 1
    while True:
        new_name = f"DDB{i}"
        if new_name not in existing_names:
            return new_name
        i += 1


def genNextStateMsgs(state_name, app_name):
    ''' Generate a set of messages based on the state name and application name '''
    params = MARTe2ConfigurationDatabase(objects = {"param1": state_name})
    prepare = MARTe2Message(f"+PrepareChangeTo{state_name}Msg", app_name,
                            "PrepareNextState", params)
    defaultmessages = [MARTe2Message("+StopCurrentStateExecutionMsg",
                                     app_name, "StopCurrentStateExecution"),
                        prepare,
                        MARTe2Message("+StartNextStateExecutionMsg",
                                      app_name, "StartNextStateExecution")]
    return defaultmessages

def showErrorDialog(message):
    ''' Show an error dialog messagebox. '''
    msg_box = QMessageBox()
    msg_box.setIcon(QMessageBox.Critical)
    msg_box.setText("Error")
    msg_box.setInformativeText(message)
    msg_box.setWindowTitle("Error")
    msg_box.exec_()

def textExistsInListWidget(list_widget, text):
    ''' Check whether list items exist in a list widget '''
    for index in range(list_widget.count()):
        item = list_widget.item(index)
        if item.text() == text:
            return True
    return False

def deleteChildren(layout):
    ''' Delete all the children in a layout '''
    # Remove all widgets from layout
    for i in reversed(range(layout.count())):
        widget = layout.itemAt(i).widget()
        if widget:
            widget.deleteLater()



'''
This module provides capability for a user to convert a MARTe2 written binary file into a human
readable CSV file.
'''

import struct
from typing import Type, Tuple, Iterable
from functools import lru_cache
import itertools
from dataclasses import dataclass


class Bin2CSVError(Exception):
    """Generic exception for all errors"""

@dataclass
class TypeDescriptor:
    ''' Basic descriptor of a type in it's binary format '''
    code: int
    name: str
    struct_format: str
    size: int

@dataclass
class SignalDescriptor:
    ''' Basic Descriptor of a signal in it's binary format '''
    name: str
    type_descr: TypeDescriptor
    elements: int

    def format(self):
        ''' Format the descriptor into string representation '''
        if self.elements == 1:
            return self.type_descr.struct_format
        return f'{self.elements:d}{self.type_descr.struct_format}'

standard_type_descriptors = {
    t.code: t for t in (
        TypeDescriptor(2048, 'int32', 'i', 4),
        TypeDescriptor(2052, 'uint32', 'I', 4),
        TypeDescriptor(2056, 'float32','f', 4),
        TypeDescriptor(4100, 'uint64', 'Q', 8),
        TypeDescriptor(4104, 'float64','d',8),
        TypeDescriptor(1028, 'uint16', 'H', 2),
        TypeDescriptor(516, 'uint8', 'B', 1),
        TypeDescriptor(512, 'int8', 'b', 1)
    )
}


class RowInterpreter:
    """Class using the descriptors built from the start of a MARTe2 binary
    FileWriter output to interpret the bytes from the rest of the file, with
    the ability to output what it interprets for storage in a CSV. If the CSV
    format is expected to be compatible outside of the MARTe2 world, see also
    the `StandardCSVRowInterpreter` class.
    """

    def __init__(self, *signal_descriptors: Tuple[SignalDescriptor],
                 delimiter: str = ',', struct_endianity_char: str = '<'):
        """RowInterpreter constructor.

        Constructor parameters are stored in private fields because the rowSize
        and rowFormat results are cached and will be invalid if these fields
        are edited after they have been called the first time.
        """
        self._signal_descriptors = signal_descriptors
        self._delimeter = delimiter
        self._struct_endianity_char = struct_endianity_char

    @lru_cache(maxsize=None) # pylint: disable=W1518
    def rowSize(self) -> int:
        """Return the number of bytes that one row of data would require in the
        MARTe2 binary file.
        """
        return sum(signal.type_descr.size * signal.elements for
                   signal in self._signal_descriptors)

    @lru_cache(maxsize=None) # pylint: disable=W1518
    def rowFormat(self) -> str:
        """Return the `struct` library format string used to unpack one row."""
        return self._struct_endianity_char + ''.join(signal.format() for
                                                     signal in self._signal_descriptors)

    def headerRow(self) -> str:
        """Return the first row of an equivalent CSV file format, containing
        column names.
        """
        column_names = [f'{signal.name} ({signal.type_descr.name})[{signal.elements}]' for
                        signal in self._signal_descriptors]
        signal_names_string = self._delimeter.join(column_names)
        return f'#{signal_names_string}\n'

    def binToCsvRow(self, binary_data: bytes) -> str:
        """Return the CSV file formatted data for a row's worth of binary data.
        This will raise an exception if the provided data does not match in
        size with the descriptors this RowInterpreter was constructed with.
        """
        return self.valuesToCsvRow(self.binToValues(binary_data))

    def binToValues(self, binary_data: bytes) -> Iterable:
        """Return the values in Python types for a row's worth of binary data.
        This will raise an exception if the provided data does not match in
        size with the descriptors this RowInterpreter was constructed with.
        """
        return struct.unpack(self.rowFormat(), binary_data)

    def valuesToCsvRow(self, row_values: Iterable) -> str:
        """Return the CSV file formatted data for a row's data as Python types.
        This will raise an exception if the provided data does not match in
        number of items with the descriptors this RowInterpreter was constructed
        with.
        """
        row_value_generator = (i for i in row_values)
        csv_value_strings = []
        signals = [(signal, itertools.islice(row_value_generator, signal.elements)) for
                   signal in self._signal_descriptors]
        for signal, signal_iterable in signals:
            if signal.elements == 1:
                csv_value_strings.append(str(next(signal_iterable)))
            else:
                inner_values = ' '.join(str(value) for value in signal_iterable)
                csv_value_strings.append(f'{{{inner_values}}}')
        return self._delimeter.join(csv_value_strings) + '\n'


class StandardCSVRowInterpreter(RowInterpreter):
    """A variant of the `RowInterpreter` class which writes a more standardised
    CSV format. The difference is that signals which are arrays in MARTe2 will
    be separated into individual signals, whereas the `RowInterpreter` class
    follows the MARTe2 convention of encasing the array values in curly braces
    and allocating them only a single header item.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def valuesToCsvRow(self, row_values: Iterable) -> str:
        return self._delimeter.join(str(value) for value in row_values) + '\n'

    def headerRow(self) -> str:
        signal_names_string = self._delimeter.join(self._headerItem(signal) for
                                                   signal in self._signal_descriptors)
        return f'#{signal_names_string}\n'

    def _headerItem(self, signal: SignalDescriptor) -> str:
        if signal.elements == 1:
            return f'{signal.name} ({signal.type_descr.name})'
        indices = range(signal.elements)
        signal_desc = f'({signal.type_descr.name}[{signal.elements}])'
        index_strings = [f'{signal.name}[{i}] {signal_desc}' for i in indices]
        return self._delimeter.join(index_strings)


class Bin2CSV():
    '''
    The main entrypoint to convert a MARTe2 binary log produced by the FileWriter datasource
    to a csv file.
    See RowInterpretator and StandardRowInterpretator for explanation of marte2_csv
    '''
    def __init__(self,
        bin_path: str = 'log.bin',
        csv_path: str = 'log_0.csv',
        marte2_csv: bool = True
    ):
        self.bin_path = bin_path
        self.csv_path = csv_path
        self.marte2_csv = marte2_csv

    def interpreterClass(self) -> Type[RowInterpreter]:
        ''' Get our declared interpretator'''
        if self.marte2_csv:
            return RowInterpreter
        return StandardCSVRowInterpreter

    @lru_cache(maxsize=None) # pylint: disable=W1518
    def unpackStruct(self, name_size, fin):
        ''' Unpack the signal '''
        return struct.unpack('<H' + str(name_size) + 'sI', fin.read((name_size+6)))

    def main(self):
        """Convert the binary file into CSV"""
        with open(self.bin_path, 'rb') as fin:
            nsignals, = struct.unpack('<I', fin.read(4))

            name_size = 32

            signals = []
            for _ in range(nsignals):
                signal_code, signal_name, signal_elements = self.unpackStruct(name_size, fin)

                signal_name = signal_name.rstrip(b'\x00').decode('utf-8')
                try:
                    signal_type = standard_type_descriptors[signal_code]
                except KeyError:
                    msg = f'Unknown TypeDescription code: {signal_code} for signal {signal_name}'
                    raise Bin2CSVError(msg) # pylint: disable=W0707

                signals.append(SignalDescriptor(signal_name, signal_type, signal_elements))

            interpreter = self.interpreterClass()(*signals)

            with open(self.csv_path, 'w', encoding='utf-8') as fout:
                fout.write(interpreter.headerRow())

                while True:
                    raw_row = fin.read(interpreter.rowSize())
                    if len(raw_row) < interpreter.rowSize():
                        if raw_row:
                            msg = f'Reached EOF with partial row: {len(raw_row)} leftover bytes'
                            raise Bin2CSVError(msg)
                        break

                    fout.write(interpreter.binToCsvRow(raw_row))



''' Pythonic representation of the SSM GAM'''
from martepy.marte2.gam import MARTe2GAM
from martepy.marte2.qt_functions import (addComboEdit,
                                         addInputSignalsSection,
                                         addLineEdit,
                                         addOutputSignalsSection)

class SSMGAM(MARTe2GAM):
    ''' Pythonic representation of the SSM GAM'''
    def __init__(self,
                    configuration_name: str = 'SSMGAM',
                    state_matrix = '{}',
                    input_matrix = '{}',
                    output_matrix = '{}',
                    feedthroughmatrix = '{}',
                    samplefreq = 1,
                    reset_each_state = 1,
                    input_signals: list = [],
                    output_signals: list = [],
                ):
        self.statematrix = state_matrix
        self.inputmatrix = input_matrix
        self.outputmatrix = output_matrix
        self.feedthroughmatrix = feedthroughmatrix
        self.samplefrequency = samplefreq
        self.resetineachstate = reset_each_state
        super().__init__(
                configuration_name = configuration_name,
                class_name = 'SSMGAM',
                input_signals = input_signals,
                output_signals = output_signals,
            )

    def writeGamConfig(self, config_writer):
        ''' Write our GAM Configuration '''
        reset_state = '0' if str(self.resetineachstate) == '0' else '1'
        config_writer.writeNode('StateMatrix', f'{self.statematrix}')
        config_writer.writeNode('InputMatrix', f'{self.inputmatrix}')
        config_writer.writeNode('OutputMatrix', f'{self.outputmatrix}')
        config_writer.writeNode('FeedthroughMatrix', f'{self.feedthroughmatrix}')
        config_writer.writeNode('ResetInEachState', reset_state)
        config_writer.writeNode('SampleFrequency', f'{self.samplefrequency}')

    def serialize(self):
        ''' Serialize the object '''
        res: dict = super().serialize()
        res['parameters']['statematrix'] = self.statematrix
        res['parameters']['inputmatrix'] = self.inputmatrix
        res['parameters']['outputmatrix'] = self.outputmatrix
        res['parameters']['feedthroughmatrix'] = self.feedthroughmatrix
        res['parameters']['resetineachstate'] = int(self.resetineachstate)
        res['parameters']['samplefrequency'] = float(self.samplefrequency)
        res['label'] = "SSMGAM"
        res['block_type'] = 'SSMGAM'
        res['class_name'] = 'SSMGAM'
        res['title'] = f"{self.configuration_name} (SSMGAM)"
        return res

    def deserialize(self, data: dict, hashmap: dict={}, restore_id: bool=True) -> bool:
        ''' Deserialize the given object to the class instance '''
        res = super().deserialize(data, hashmap, restore_id)
        self.statematrix = data['parameters']["statematrix"]
        self.inputmatrix = data['parameters']["inputmatrix"]
        self.outputmatrix = data['parameters']["outputmatrix"]
        self.feedthroughmatrix = data['parameters']["feedthroughmatrix"]
        self.resetineachstate = int(data['parameters']["resetineachstate"])
        self.samplefrequency = data['parameters']["samplefrequency"]
        return res

    @staticmethod
    def loadParameters(mainpanel_instance, node):
        """This function is intended to be for the GUI where it can
        call the static instance of the class directly to generate
        the appropriate parameter modifier for the node in XMARTe2.
        """
        app_def = mainpanel_instance.parent.API.getServiceByName('ApplicationDefinition')
        datasource = app_def.configuration['misc']['gamsources'][0]
        addInputSignalsSection(mainpanel_instance, node, False, True)

        addOutputSignalsSection(mainpanel_instance, node, 3, False, datasource, True)

        addLineEdit(mainpanel_instance, node, "StateMatrix: ", 'statematrix', 3, 0)

        addLineEdit(mainpanel_instance, node, "InputMatrix: ", 'inputmatrix', 3, 2)

        addLineEdit(mainpanel_instance, node, "OutputMatrix: ", 'outputmatrix', 4, 0)

        addLineEdit(mainpanel_instance, node, "Feed through Matrix: ", 'feedthroughmatrix', 4, 2)

        addComboEdit(mainpanel_instance, node, "Reset in each State:",
                     "resetineachstate", 5, 0, ['1', '0'])

        addLineEdit(mainpanel_instance, node, "Sample Frequency: ", 'samplefrequency', 5, 2)

def initialize(factory, plugin_datastore) -> None:
    ''' Initialize the object with the factory '''
    factory.registerBlock("SSMGAM", SSMGAM, plugin_datastore)
